use core::fmt::Debug;
use std::{collections::VecDeque, fmt::{Display, Result}, sync::{Arc, Mutex}};

use rgb::Rgb;

#[derive(Debug, Clone, PartialEq, Eq)]
pub enum Variant {
    SignedByte(i8),
    Byte(u8),
    UInt(u32),
    Int(i32),
    BigUInt(u64),
    BigInt(i64),
    Boolean(bool),
    String(String),
    RGB(Rgb<u8>),
    Vec(Vec<Variant>)
}

macro_rules! impl_variant_type {
    ($type:ty, $var_type:tt) => {
        impl From<$type> for Variant {
            fn from(value: $type) -> Self {
                Variant::$var_type(value)
            }
        }

        impl Into<$type> for Variant {
            fn into(self) -> $type {
                match self {
                    Variant::$var_type(value) => value,
                    _ => panic!(concat!("Expected Variant::", stringify!($var_type), "but got {:?}"), self)
                }
            }
        }
    };
}

impl_variant_type!(u8, Byte);
impl_variant_type!(i8, SignedByte);
impl_variant_type!(u32, UInt);
impl_variant_type!(i32, Int);
impl_variant_type!(i64, BigInt);
impl_variant_type!(u64, BigUInt);
impl_variant_type!(bool, Boolean);
impl_variant_type!(String, String);
impl_variant_type!(Rgb<u8>, RGB);
impl_variant_type!(Vec<Variant>, Vec);

impl<'a> From<&'a str> for Variant {
    fn from(value: &'a str) -> Self {
        Variant::String(value.to_owned())
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub enum Event {
    ReadyToRock,
    Tick,
    StartThing(&'static str),
    StopThing(&'static str),
    PropertyChange(PropertyID, Variant)
}

impl Display for Event {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result {
        match self {
            Self::ReadyToRock => f.write_str("ReadyToRock"),
            Self::Tick => f.write_str("Tick"),
            Self::StartThing(name) => write!(f, "Start {}", *name),
            Self::StopThing(name) => write!(f, "Stop {}", *name),
            Self::PropertyChange(id, value) => write!(f, "{} -> {:?}", id, value)
        }
    }
}

impl Event {
    pub const fn new_tick() -> Self {
        Event::Tick
    }

    pub fn new_property_change<T>(key: PropertyID, data: T) -> Self where Variant: From<T> {
        Event::PropertyChange(key, Variant::from(data))
    }

    pub const fn new_ready_to_rock() -> Self {
        Event::ReadyToRock
    }

    pub const fn new_start_thing(name: &'static str) -> Self {
        Event::StartThing(name)
    }

    pub const fn new_stop_thing(name: &'static str) -> Self {
        Event::StopThing(name)
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Property {
    key: PropertyKey,
    value: Variant
}

impl Property {
    pub const fn new(key: PropertyKey, value: Variant) -> Self {
        Property {
            key,
            value,
        }
    }
}

#[derive(Debug, Clone, Copy, Eq, PartialEq)]
pub struct NamespaceKey(pub &'static str);

#[derive(Debug, Clone, Copy, Eq, PartialEq)]
pub struct PropertyKey(pub &'static str);

#[derive(Debug, Clone, Copy, Eq, PartialEq)]
pub struct PropertyID {
    pub namespace: NamespaceKey,
    pub key: PropertyKey
}

impl Display for PropertyID {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result {
        write!(f, "{}::{}", self.namespace.0, self.key.0)
    }
}

pub trait Namespace {
    fn nskey() -> NamespaceKey;
    fn properties() -> Vec<Property>;
}

#[macro_export]
macro_rules! prop_id {
    ($name:ident, $prop_name:ident) => {
        $crate::PropertyID {
            namespace: $crate::NamespaceKey(stringify!($name)),
            key: $crate::PropertyKey(stringify!($prop_name))
        }
    };
}

#[macro_export]
macro_rules! property_namespace {
    ($name:ident, $($prop_name:ident => $value:expr),*) => {
        pub enum $name {
            $($prop_name),*
        }

        impl $crate::Namespace for $name {
            fn nskey() -> $crate::NamespaceKey {
                $crate::NamespaceKey(stringify!($name))
            }

            fn properties() -> Vec<$crate::Property> {
                vec![
                    $($crate::Property::new(Self::$prop_name.key(), $crate::Variant::from($value))),*
                ]
            }
        }

        impl $name {
            pub const fn key(&self) -> $crate::PropertyKey {
                match self {
                    $(Self::$prop_name => $crate::PropertyKey(stringify!($prop_name))),*
                }
            }

            pub const fn id(&self) -> $crate::PropertyID {
                $crate::PropertyID { namespace: $crate::NamespaceKey(stringify!($name)), key: self.key() }
            }

            pub fn new_property_change<V>(&self, value: V) -> $crate::Event where $crate::Variant: From<V> {
                $crate::Event::PropertyChange(self.id(), $crate::Variant::from(value))
            }
        }

        impl Into<$crate::PropertyID> for $name {
            fn into(self) -> $crate::PropertyID {
                self.id()
            }
        }
    };
}

#[derive(Debug, Clone)]
pub struct NamespaceProps {
    props: Vec<Property>,
    key: NamespaceKey
}

impl NamespaceProps {
    fn new<NS: Namespace>() -> Self {
        NamespaceProps {
            props: NS::properties(),
            key: NS::nskey()
        }
    }

    fn get_key(&self, key: &PropertyKey) -> Option<&Property> {
        for next in self.props.iter() {
            if next.key == *key {
                return Some(next);
            }
        }

        log::warn!("Unknown key {:?}", key);

        return None;
    }

    fn get_key_mut(&mut self, key: &PropertyKey) -> Option<&mut Property> {
        for next in self.props.iter_mut() {
            if next.key == *key {
                return Some(next);
            }
        }

        log::warn!("Unknown key {:?}", key);

        return None;
    }

}

#[derive(Debug, Clone)]
pub struct Properties {
    contents: Vec<NamespaceProps>
}

impl Display for Properties {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        for ns in self.contents.iter() {
            write!(f, "{}\n", ns.key.0).unwrap();
            for prop in ns.props.iter() {
                write!(f, "\t{} = {:?}\n", prop.key.0, prop.value).unwrap();
            }
        }
        Result::Ok(())
    }
}

impl Properties {
    fn new() -> Self {
        Properties {
            contents: Vec::new(),
        }
    }

    pub fn add_namespace<T: Namespace>(&mut self) {
        self.contents.push(NamespaceProps::new::<T>());
    }

    fn get_namespace(&self, ns: &NamespaceKey) -> Option<&NamespaceProps> {
        for nsprops in self.contents.iter() {
            if nsprops.key == *ns {
                return Some(nsprops)
            }
        }

        log::warn!("Unknown namespace {:?}", ns);

        None
    }

    fn get_namespace_mut(&mut self, ns: &NamespaceKey) -> Option<&mut NamespaceProps> {
        for nsprops in self.contents.iter_mut() {
            if nsprops.key == *ns {
                return Some(nsprops)
            }
        }

        log::warn!("Unknown namespace {:?}", ns);

        None
    }

    pub fn get<T: Into<PropertyID>>(&self, key: T) -> Option<Variant> {
        let as_id = key.into();
        match self.get_namespace(&as_id.namespace).unwrap().get_key(&as_id.key) {
            None => None,
            Some(v) => Some(v.value.clone())
        }
    }

    pub fn set<V, T: Into<PropertyID>>(&mut self, key: T, value: V) -> bool where Variant: From<V> {
        let as_id = key.into();
        let as_variant: Variant = value.into();
        match self.get_namespace_mut(&as_id.namespace).unwrap().get_key_mut(&as_id.key) {
            None => (),
            Some(found_key) => {
                if found_key.value != as_variant {
                    found_key.value = as_variant;
                    return true
                }
            }
        }

        return false
    }
}

#[derive(Clone)]
pub struct EventBus {
    pending: Arc<Mutex<VecDeque<Event>>>,
    props: Arc<Mutex<Properties>>
}

impl EventBus {
    pub fn new() -> Self {
        EventBus {
            pending: Arc::new(Mutex::new(VecDeque::with_capacity(32))),
            props: Arc::new(Mutex::new(Properties::new()))
        }
    }

    pub fn next(&mut self) -> Event {
        let next = self.pending.lock().unwrap().pop_front().unwrap_or(Event::new_tick());

        match next {
            Event::PropertyChange(key, value) => {
                if self.props.lock().unwrap().set(key, value.clone()) {
                    log::trace!("prop-update key={:?} value={:?}", key, value);
                    Event::PropertyChange(key, value)
                } else {
                    Event::new_tick()
                }
            },
            _ => next
        }
    }

    pub fn push(&mut self, event: Event) {
        self.pending.lock().unwrap().push_back(event);
    }

    pub fn properties(&self) -> std::sync::MutexGuard<'_, Properties> {
        self.props.lock().unwrap()
    }

    pub fn set_property<K: Into<PropertyID>, V: Into<Variant>>(&mut self, key: K, value: V) {
        self.push(Event::new_property_change(key.into(), value.into()));
    }
}

property_namespace!(
    System,
    NetworkOnline => false,
    IP => "",
    Gateway => "",
    TimeSync => false
);